#define _GNU_SOURCE // required by getline

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/time.h>
#include <math.h>
#include <pthread.h>

#include <CL/cl.h>

#ifdef _STATIC_MAX_PLATFORMS_AND_DEVICES
#define MAX_PLATFORMS 16
#define MAX_DEVICES 4
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#define MATRIX_FILE "cl_matrix.out"

#include <errno.h>	// required by errnum

float *GLOBAL_A, *GLOBAL_B, *GLOBAL_C; 

int GLOBAL_size, GLOBAL_slice, GLOBAL_tile;
char **GLOBAL_kernel_source_code;

#ifdef _STATIC_MAX_PLATFORMS_AND_DEVICES
struct{
	cl_platform_id platform;
	cl_device_id device;
	size_t resolution;
	size_t total_slices;
	size_t total_enq_write_bytes;
	size_t total_enq_read_bytes;
	cl_ulong total_kernel_work;
} GLOBAL_dev_info[MAX_PLATFORMS*MAX_DEVICES];
pthread_t GLOBAL_dev_thread[MAX_PLATFORMS*MAX_DEVICES];
#else
typedef struct{
	cl_platform_id platform;
	cl_device_id device;
	size_t resolution;
	size_t total_slices;
	size_t total_enq_write_bytes;
	size_t total_enq_read_bytes;
	cl_ulong total_kernel_work;
} dev_info_t;
dev_info_t *GLOBAL_dev_info=NULL;
pthread_t *GLOBAL_dev_thread=NULL;
#endif

cl_uint GLOBAL_num_devices=0;
pthread_mutex_t GLOBAL_get_work_mutex=PTHREAD_MUTEX_INITIALIZER;
pthread_barrier_t GLOBAL_ccl_barrier;

////////////////////////////////////////////////////////////////////////////////

void cclTestSuccess(char *where, cl_int error){
	char error_name[65][44]={"CL_SUCCESS", "CL_DEVICE_NOT_FOUND", "CL_DEVICE_NOT_AVAILABLE", "CL_COMPILER_NOT_AVAILABLE", "CL_MEM_OBJECT_ALLOCATION_FAILURE", "CL_OUT_OF_RESOURCES", "CL_OUT_OF_HOST_MEMORY", "CL_PROFILING_INFO_NOT_AVAILABLE", "CL_MEM_COPY_OVERLAP", "CL_IMAGE_FORMAT_MISMATCH", "CL_IMAGE_FORMAT_NOT_SUPPORTED", "CL_BUILD_PROGRAM_FAILURE", "CL_MAP_FAILURE", "CL_MISALIGNED_SUB_BUFFER_OFFSET", "CL_EXEC_STATUS_ERROR_FOR_EVENTS_IN_WAIT_LIST", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "CL_INVALID_VALUE", "CL_INVALID_DEVICE_TYPE", "CL_INVALID_PLATFORM", "CL_INVALID_DEVICE", "CL_INVALID_CONTEXT",  "CL_INVALID_QUEUE_PROPERTIES", "CL_INVALID_COMMAND_QUEUE", "CL_INVALID_HOST_PTR", "CL_INVALID_MEM_OBJECT", "CL_INVALID_IMAGE_FORMAT_DESCRIPTOR", "CL_INVALID_IMAGE_size", "CL_INVALID_SAMPLER", "CL_INVALID_BINARY", "CL_INVALID_BUILD_OPTIONS", "CL_INVALID_PROGRAM", "CL_INVALID_PROGRAM_EXECUTABLE", "CL_INVALID_KERNEL_NAME", "CL_INVALID_KERNEL_DEFINITION", "CL_INVALID_KERNEL", "CL_INVALID_ARG_INDEX", "CL_INVALID_ARG_VALUE", "CL_INVALID_ARG_size", "CL_INVALID_KERNEL_ARGS", "CL_INVALID_WORK_DIMENSION", "CL_INVALID_WORK_GROUP_size", "CL_INVALID_WORK_ITEM_size", "CL_INVALID_GLOBAL_OFFSET", "CL_INVALID_EVENT_WAIT_LIST", "CL_INVALID_EVENT", "CL_INVALID_OPERATION", "CL_INVALID_GL_OBJECT", "CL_INVALID_BUFFER_size", "CL_INVALID_MIP_LEVEL", "CL_INVALID_GLOBAL_WORK_size", "CL_INVALID_PROPERTY"};

	if(error != CL_SUCCESS){
		printf("%s: %s\n", where, error_name[-error]);
		exit(1);
	}
}

////////////////////////////////////////////////////////////////////////////////

void cclPerror(char *where, cl_int error){
	char error_name[65][44]={"CL_SUCCESS", "CL_DEVICE_NOT_FOUND", "CL_DEVICE_NOT_AVAILABLE", "CL_COMPILER_NOT_AVAILABLE", "CL_MEM_OBJECT_ALLOCATION_FAILURE", "CL_OUT_OF_RESOURCES", "CL_OUT_OF_HOST_MEMORY", "CL_PROFILING_INFO_NOT_AVAILABLE", "CL_MEM_COPY_OVERLAP", "CL_IMAGE_FORMAT_MISMATCH", "CL_IMAGE_FORMAT_NOT_SUPPORTED", "CL_BUILD_PROGRAM_FAILURE", "CL_MAP_FAILURE", "CL_MISALIGNED_SUB_BUFFER_OFFSET", "CL_EXEC_STATUS_ERROR_FOR_EVENTS_IN_WAIT_LIST", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "CL_INVALID_VALUE", "CL_INVALID_DEVICE_TYPE", "CL_INVALID_PLATFORM", "CL_INVALID_DEVICE", "CL_INVALID_CONTEXT",  "CL_INVALID_QUEUE_PROPERTIES", "CL_INVALID_COMMAND_QUEUE", "CL_INVALID_HOST_PTR", "CL_INVALID_MEM_OBJECT", "CL_INVALID_IMAGE_FORMAT_DESCRIPTOR", "CL_INVALID_IMAGE_SIZE", "CL_INVALID_SAMPLER", "CL_INVALID_BINARY", "CL_INVALID_BUILD_OPTIONS", "CL_INVALID_PROGRAM", "CL_INVALID_PROGRAM_EXECUTABLE", "CL_INVALID_KERNEL_NAME", "CL_INVALID_KERNEL_DEFINITION", "CL_INVALID_KERNEL", "CL_INVALID_ARG_INDEX", "CL_INVALID_ARG_VALUE", "CL_INVALID_ARG_SIZE", "CL_INVALID_KERNEL_ARGS", "CL_INVALID_WORK_DIMENSION", "CL_INVALID_WORK_GROUP_SIZE", "CL_INVALID_WORK_ITEM_SIZE", "CL_INVALID_GLOBAL_OFFSET", "CL_INVALID_EVENT_WAIT_LIST", "CL_INVALID_EVENT", "CL_INVALID_OPERATION", "CL_INVALID_GL_OBJECT", "CL_INVALID_BUFFER_SIZE", "CL_INVALID_MIP_LEVEL", "CL_INVALID_GLOBAL_WORK_SIZE", "CL_INVALID_PROPERTY"};
	
	printf("%s: %s\n", where, error_name[-error]);
	exit(1);
}

////////////////////////////////////////////////////////////////////////////////

void cclTestSuccess2(char *where, cl_int error, int line){
	char error_name[65][44]={"CL_SUCCESS", "CL_DEVICE_NOT_FOUND", "CL_DEVICE_NOT_AVAILABLE", "CL_COMPILER_NOT_AVAILABLE", "CL_MEM_OBJECT_ALLOCATION_FAILURE", "CL_OUT_OF_RESOURCES", "CL_OUT_OF_HOST_MEMORY", "CL_PROFILING_INFO_NOT_AVAILABLE", "CL_MEM_COPY_OVERLAP", "CL_IMAGE_FORMAT_MISMATCH", "CL_IMAGE_FORMAT_NOT_SUPPORTED", "CL_BUILD_PROGRAM_FAILURE", "CL_MAP_FAILURE", "CL_MISALIGNED_SUB_BUFFER_OFFSET", "CL_EXEC_STATUS_ERROR_FOR_EVENTS_IN_WAIT_LIST", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "CL_INVALID_VALUE", "CL_INVALID_DEVICE_TYPE", "CL_INVALID_PLATFORM", "CL_INVALID_DEVICE", "CL_INVALID_CONTEXT",  "CL_INVALID_QUEUE_PROPERTIES", "CL_INVALID_COMMAND_QUEUE", "CL_INVALID_HOST_PTR", "CL_INVALID_MEM_OBJECT", "CL_INVALID_IMAGE_FORMAT_DESCRIPTOR", "CL_INVALID_IMAGE_size", "CL_INVALID_SAMPLER", "CL_INVALID_BINARY", "CL_INVALID_BUILD_OPTIONS", "CL_INVALID_PROGRAM", "CL_INVALID_PROGRAM_EXECUTABLE", "CL_INVALID_KERNEL_NAME", "CL_INVALID_KERNEL_DEFINITION", "CL_INVALID_KERNEL", "CL_INVALID_ARG_INDEX", "CL_INVALID_ARG_VALUE", "CL_INVALID_ARG_size", "CL_INVALID_KERNEL_ARGS", "CL_INVALID_WORK_DIMENSION", "CL_INVALID_WORK_GROUP_size", "CL_INVALID_WORK_ITEM_size", "CL_INVALID_GLOBAL_OFFSET", "CL_INVALID_EVENT_WAIT_LIST", "CL_INVALID_EVENT", "CL_INVALID_OPERATION", "CL_INVALID_GL_OBJECT", "CL_INVALID_BUFFER_size", "CL_INVALID_MIP_LEVEL", "CL_INVALID_GLOBAL_WORK_size", "CL_INVALID_PROPERTY"};

	if(error != CL_SUCCESS){
		printf("line %d: %s: %s\n", line, where, error_name[-error]);
		exit(1);
	}
}

////////////////////////////////////////////////////////////////////////////////

void cclPerror2(char *where, cl_int error, int line)
{
	char error_name[65][44]={"CL_SUCCESS", "CL_DEVICE_NOT_FOUND", "CL_DEVICE_NOT_AVAILABLE", "CL_COMPILER_NOT_AVAILABLE", "CL_MEM_OBJECT_ALLOCATION_FAILURE", "CL_OUT_OF_RESOURCES", "CL_OUT_OF_HOST_MEMORY", "CL_PROFILING_INFO_NOT_AVAILABLE", "CL_MEM_COPY_OVERLAP", "CL_IMAGE_FORMAT_MISMATCH", "CL_IMAGE_FORMAT_NOT_SUPPORTED", "CL_BUILD_PROGRAM_FAILURE", "CL_MAP_FAILURE", "CL_MISALIGNED_SUB_BUFFER_OFFSET", "CL_EXEC_STATUS_ERROR_FOR_EVENTS_IN_WAIT_LIST", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "CL_INVALID_VALUE", "CL_INVALID_DEVICE_TYPE", "CL_INVALID_PLATFORM", "CL_INVALID_DEVICE", "CL_INVALID_CONTEXT",  "CL_INVALID_QUEUE_PROPERTIES", "CL_INVALID_COMMAND_QUEUE", "CL_INVALID_HOST_PTR", "CL_INVALID_MEM_OBJECT", "CL_INVALID_IMAGE_FORMAT_DESCRIPTOR", "CL_INVALID_IMAGE_SIZE", "CL_INVALID_SAMPLER", "CL_INVALID_BINARY", "CL_INVALID_BUILD_OPTIONS", "CL_INVALID_PROGRAM", "CL_INVALID_PROGRAM_EXECUTABLE", "CL_INVALID_KERNEL_NAME", "CL_INVALID_KERNEL_DEFINITION", "CL_INVALID_KERNEL", "CL_INVALID_ARG_INDEX", "CL_INVALID_ARG_VALUE", "CL_INVALID_ARG_SIZE", "CL_INVALID_KERNEL_ARGS", "CL_INVALID_WORK_DIMENSION", "CL_INVALID_WORK_GROUP_SIZE", "CL_INVALID_WORK_ITEM_SIZE", "CL_INVALID_GLOBAL_OFFSET", "CL_INVALID_EVENT_WAIT_LIST", "CL_INVALID_EVENT", "CL_INVALID_OPERATION", "CL_INVALID_GL_OBJECT", "CL_INVALID_BUFFER_SIZE", "CL_INVALID_MIP_LEVEL", "CL_INVALID_GLOBAL_WORK_SIZE", "CL_INVALID_PROPERTY"};
	
	printf("line %d: %s: %s\n", line, where, error_name[-error]);
	exit(1);
}

////////////////////////////////////////////////////////////////////////////////

void cclPerror3(const char *primitive, int line)
{
	printf("line %d: %s: %s\n", line, primitive, strerror(errno));
	exit(errno);
}

////////////////////////////////////////////////////////////////////////////////

void cclShowArgv(int argc, char **argv)
{
	int i;

	for (i=0; i<argc; i++)
		printf("argv[%d]=%s\n", i, argv[i]);
	
}

////////////////////////////////////////////////////////////////////////////////

void cclPrintUsage(char **argv)
{		
		printf("Usage: %s\n\t-q [ --txt | --env | --file ] |", argv[0]);
		printf("\n\t-S size -s slice -t tile { -f file |\n");
		printf("\t                           -p all |\n");
		printf("\t                           -p platform_index -d device_index [ -p platform_index -d device_index ... ] }\n");
}

////////////////////////////////////////////////////////////////////////////////

void cclPrintMatrix(float *M, int size){
	int i, j;
	
	for(i=0; i<size; i++){
		for(j=0; j<size; j++)
			printf("%4.0f ", M[i*size+j]);
		printf("\n");
	}
	printf("\n");
}

////////////////////////////////////////////////////////////////////////////////


void cclSaveMatrix(float *M, int size, const char *file){
	int fd, ret;

	printf("cclSaveMatrix: saving %lu floats (%lu bytes)\n", (long unsigned int)(size*size), (long unsigned int)(size*size)*sizeof(float));
	fd=creat(file, 00700);
	if (fd < 0)
		cclPerror3("creat", __LINE__);
	ret=write(fd, (void*)M, size*size*sizeof(float));
	if (ret < 0)
		cclPerror3("write", __LINE__);
	if (ret != size*size*sizeof(float)) {
		printf("cclSaveMatrix: ret != %lu\n", size*size*sizeof(float));
		exit(1);
	}
	close(fd);
}

////////////////////////////////////////////////////////////////////////////////

void cclCheckMatrix()
{
	int i, j;
	float v;	
	long double acc_errors=0;
	
	for(i=0; i<GLOBAL_size; i+=64){
		v = 0;
		for(j=0; j<GLOBAL_size; j++)
			v += GLOBAL_A[i*GLOBAL_size+j] * GLOBAL_B[i*GLOBAL_size+j];
			if(GLOBAL_C[i*GLOBAL_size+i] != v)
				acc_errors += fabs(GLOBAL_C[i*GLOBAL_size+i] - v);
				//printf("warning: C[%d][%d]: %f != %f\n", i, i, GLOBAL_C[i*GLOBAL_size+i], v);
        }
	printf("acc_errors=%LG\n", acc_errors);
}

////////////////////////////////////////////////////////////////////////////////

char ** cclReadProgramSources(int nfiles, ...){
	FILE *file;
	long size;
	va_list ap;
	int n;
	char **strings;

	strings = malloc(sizeof(char*) * nfiles);
	va_start(ap, nfiles);
	for(n=0; n<nfiles; n++){
		file = fopen(va_arg(ap, char*), "r");
		fseek(file, 0, SEEK_END);
		size = ftell(file);
		rewind(file);
		(strings)[n] = (char *)malloc(sizeof(char) * size + 1);
		fread((strings)[n], 1, sizeof(char) * size, file);
		(strings)[n][size] = '\0';
		fclose(file);
	}
	va_end(ap);

	return(strings);
}

////////////////////////////////////////////////////////////////////////////////

void cclQueryPlatformsAndDevices(int argc, char **argv)
{
	// IN:
	// 		argv: ./cl_matrix -q [ --txt | --env | --file ]
	
	// Use this to check the output of each API call
	cl_int status;
	
	//---------------------------------------------------
	// Discover the platforms
	//---------------------------------------------------

	cl_uint numPlatforms = 0;
		
#ifdef _STATIC_MAX_PLATFORMS_AND_DEVICES
	cl_platform_id platforms[MAX_PLATFORMS];
	
	status = clGetPlatformIDs(MAX_PLATFORMS, platforms, &numPlatforms);
	cclTestSuccess2("clGetPlatformIDs", status, __LINE__);
#else
	cl_platform_id *platforms = NULL;

	// Use clGetPlatformIDs() to retrieve the number of platforms
	status = clGetPlatformIDs(0, NULL, &numPlatforms);
	cclTestSuccess2("clGetPlatformIDs", status, __LINE__);

	// Allocate enough space for each platform
	platforms = (cl_platform_id*)malloc(numPlatforms*sizeof(cl_platform_id));

	// Fill in platforms with clGetPlatformIDs()
	status = clGetPlatformIDs(numPlatforms, platforms, NULL);
	cclTestSuccess2("clGetPlatformIDs", status, __LINE__);
#endif

	//---------------------------------------------------
	// For each platorm, show platform and discover (and show) its devices
	//---------------------------------------------------
	int p;
	for (p=0; p < numPlatforms; p++) {
		char param_value[255]="localhost"; size_t param_value_size_ret;

		status=clGetPlatformInfo(platforms[p], 0x000, 255, (void*)param_value, &param_value_size_ret);
		//cclTestSuccess2("clGetPlatformInfo", status, __LINE__);		
		if (argc == 2 || !strcmp(argv[2], "--txt"))
			printf("[%s]\n", param_value);

		if (argc == 2 || !strcmp(argv[2], "--txt"))
			printf("\t[-p %d] ", p);			
			
		status=clGetPlatformInfo(platforms[p], CL_PLATFORM_NAME, 255, (void*)param_value, &param_value_size_ret);
		cclTestSuccess2("clGetPlatformInfo", status, __LINE__);		
		if (argc == 2 || !strcmp(argv[2], "--txt"))
			printf("%s\n", param_value);

		cl_uint numDevices = 0;
		
#ifdef _STATIC_MAX_PLATFORMS_AND_DEVICES
		cl_device_id devices[MAX_DEVICES];

		status = clGetDeviceIDs(platforms[p], CL_DEVICE_TYPE_ALL, MAX_DEVICES, devices, &numDevices);
		cclTestSuccess2("clGetDeviceIDs", status, __LINE__);
#else			
		cl_device_id *devices = NULL;

		// Use clGetDeviceIDs() to retrieve the number of devices present
		status = clGetDeviceIDs(platforms[p], CL_DEVICE_TYPE_ALL, 0, NULL, &numDevices);
		cclTestSuccess2("clGetDeviceIDs", status, __LINE__);
						
		// Allocate enough space for each device
		devices = (cl_device_id*)malloc(numDevices*sizeof(cl_device_id));

		// Fill in devices with clGetDeviceIDs()
		status = clGetDeviceIDs(platforms[p], CL_DEVICE_TYPE_ALL, numDevices, devices, NULL);
		cclTestSuccess2("clGetDeviceIDs", status, __LINE__);
#endif

		int d;
		for (d=0; d < numDevices; d++) {
			char param_value[255]; size_t param_value_size_ret;

			if (argc == 2 || !strcmp(argv[2], "--txt"))
				printf("\t\t[-d %d] ", d);
			else if (!strcmp(argv[2], "--env"))
				printf("-p %d -d %d ", p, d);
			else if (!strcmp(argv[2], "--file"))
				printf("-p %d -d %d\n", p, d);

			status=clGetDeviceInfo(devices[d], CL_DEVICE_NAME, 255, (void*)param_value, &param_value_size_ret);
			cclTestSuccess2("clGetDeviceInfo", status, __LINE__);
			if (argc == 2 || !strcmp(argv[2], "--txt"))	
				printf("%s\n", param_value);
		}
#ifndef _STATIC_MAX_PLATFORMS_AND_DEVICES
		free(devices);
#endif
	}

#ifndef _STATIC_MAX_PLATFORMS_AND_DEVICES
	free(platforms);
#endif	
	if (argc == 3 && !strcmp(argv[2], "--env"))
		printf("\n");
}

////////////////////////////////////////////////////////////////////////////////

void cclReadPlatformsAndDevicesFromFile(int *argc_file, char ***argv_file, const char *file_path)
{
	// IN:
	//	file_path
	// OUT:
	//	argc_file (how many pairs <platform, devices> where read from file
	//	argv_file (argv-like array, storing all the pairs)
	
	FILE *fp;
	char *line;
	size_t len;
	ssize_t ret;
	
	*argv_file = NULL;
	*argc_file = 0;
	
	fp = fopen(file_path, "r");
	if (fp == NULL)
		cclPerror3("fopen", __LINE__);
	
	line=NULL; // make sure getline will malloc the line
	while( (ret = getline(&line, &len, fp)) != -1 ) {
			if (line[strlen(line)-1] == '\n')
				line[strlen(line)-1] = '\0';
			//printf("getline: line: %s\n", line);
			
			// tokenize the line
			char *str=line, *token=NULL;
			token=strtok(str," ");
			while (token != NULL) {
				//printf("strtok: token: %s\n", token);
				
				*argv_file = (char **) realloc( *argv_file, (*argc_file+1)*sizeof(char *));
				(*argv_file)[*argc_file] = strdup(token);
				(*argc_file) ++;
				
				token=strtok(NULL," ");
			}
			
			free(line);
			line = NULL; // make sure getline will malloc the line
	}
	
	//printf("argc_file: %d\n", *argc_file);
	//exit(0);
	
}
						
////////////////////////////////////////////////////////////////////////////////

void cclInitPlatformsAndDevices(int argc, char **argv)
{	
	// IN:
	// 	argc: number of strings on argv
	// 	argv: "-p","platform_index", "-d", "device_index" [ , "-p","platform_index", "-d", "device_index", ...]
	// OUT:
	//	num_devices
	//  dev_info
	
	// Use this to check the output of each API call
	cl_int status;
	
	//---------------------------------------------------
	// Discover and initialize the platforms
	//---------------------------------------------------
	
	cl_uint numPlatforms = 0;
	
#ifdef _STATIC_MAX_PLATFORMS_AND_DEVICES
	cl_platform_id platforms[MAX_PLATFORMS];
	
	status = clGetPlatformIDs(MAX_PLATFORMS, platforms, &numPlatforms);
	cclTestSuccess2("clGetPlatformIDs", status, __LINE__);
#else
	cl_platform_id *platforms = NULL;
	
	// Use clGetPlatformIDs() to retrieve the number of platforms
	status = clGetPlatformIDs(0, NULL, &numPlatforms);
	cclTestSuccess2("clGetPlatformIDs", status, __LINE__);
		
	// Allocate enough space for each platform
	platforms = (cl_platform_id*)malloc(numPlatforms*sizeof(cl_platform_id));
	
	// Fill in platforms with clGetPlatformIDs()
	status = clGetPlatformIDs(numPlatforms, platforms, NULL);
	cclTestSuccess2("clGetPlatformIDs", status, __LINE__);
#endif
	
	// for each selected platform, get the selected device
	
	GLOBAL_num_devices = 0;
#ifndef _STATIC_MAX_PLATFORMS_AND_DEVICES
	GLOBAL_dev_info = NULL;
#endif
	
	int argv_index=0;
	while (argv_index+3 <= argc-1) {
		
		int p = atoi(argv[argv_index+1]); 
		int d = atoi(argv[argv_index+3]);
		
		if (p<0 || p >= numPlatforms) {
			printf("line %d: platform %d out of range {0,...,%d}\n", __LINE__, p, numPlatforms-1);
			//free(platforms);
			exit(1);
		}
		
		char param_value[255]="localhost"; size_t param_value_size_ret;
		
		status=clGetPlatformInfo(platforms[p], 0x000, 255, (void*)param_value, &param_value_size_ret);
		//cclTestSuccess2("clGetPlatformInfo", status, __LINE__);		
		printf("host\t%s\n", param_value);
		
		printf("\t\tplatform %d\n", p);	

		status=clGetPlatformInfo(platforms[p], CL_PLATFORM_NAME, 255, (void*)param_value, &param_value_size_ret);
		cclTestSuccess2("clGetPlatformInfo", status, __LINE__);
		printf("\t\t\tCL_PLATFORM_NAME: %s\n", param_value);
	
#ifndef _STATIC_MAX_PLATFORMS_AND_DEVICES		
		GLOBAL_dev_info=(dev_info_t*)realloc(GLOBAL_dev_info, (GLOBAL_num_devices+1)*sizeof(dev_info_t));
#endif
		(GLOBAL_dev_info[GLOBAL_num_devices]).platform=platforms[p];
	
		//---------------------------------------------------
		// Discover and initialize the devices
		//---------------------------------------------------
	
		cl_uint numDevices = 0;
		
#ifdef _STATIC_MAX_PLATFORMS_AND_DEVICES
		cl_device_id devices[MAX_DEVICES];

		status = clGetDeviceIDs(platforms[p], CL_DEVICE_TYPE_ALL, MAX_DEVICES, devices, &numDevices);
		cclTestSuccess2("clGetDeviceIDs", status, __LINE__);				
#else
		cl_device_id *devices = NULL;
	
		// Use clGetDeviceIDs() to retrieve the number of devices present
		status = clGetDeviceIDs(platforms[p], CL_DEVICE_TYPE_ALL, 0, NULL, &numDevices);
		cclTestSuccess2("clGetDeviceIDs", status, __LINE__);	

		// Allocate enough space for each device
		devices = (cl_device_id*)malloc(numDevices*sizeof(cl_device_id));
		
		// Fill in devices with clGetDeviceIDs()
		status = clGetDeviceIDs(platforms[p], CL_DEVICE_TYPE_ALL, numDevices, devices, NULL);
		cclTestSuccess2("clGetDeviceIDs", status, __LINE__);
#endif
	
		if (d<0 || d >= numDevices) {
			printf("line %d: device %d out of range {0,...,%d}\n", __LINE__, d, numDevices-1);
			//free(platforms);
			//free(devices);
			exit(1);
		}
		printf("\t\t\tdevice %d\n", d);

		status=clGetDeviceInfo(devices[d], CL_DEVICE_NAME, 255, (void*)param_value, &param_value_size_ret);
		cclTestSuccess2("clGetDeviceInfo", status, __LINE__);
		printf("\t\t\t\tCL_DEVICE_NAME: %s\n", param_value);		

		(GLOBAL_dev_info[GLOBAL_num_devices]).device=devices[d];

#ifndef _STATIC_MAX_PLATFORMS_AND_DEVICES	
		free(devices);
#endif
		
		GLOBAL_num_devices++;
		argv_index += 4;
	}

#ifndef _STATIC_MAX_PLATFORMS_AND_DEVICES			
	free(platforms);
#endif

}

////////////////////////////////////////////////////////////////////////////////

void cclInitAllPlatformsAndDevices()
{	
	// OUT:
	//	num_devices
	//  dev_info
	
	// Use this to check the output of each API call
	cl_int status;
	
	//---------------------------------------------------
	// Discover and initialize the platforms
	//---------------------------------------------------
	
	cl_uint numPlatforms = 0;
	
#ifdef _STATIC_MAX_PLATFORMS_AND_DEVICES
	cl_platform_id platforms[MAX_PLATFORMS];
	
	status = clGetPlatformIDs(MAX_PLATFORMS, platforms, &numPlatforms);
	cclTestSuccess2("clGetPlatformIDs", status, __LINE__);
#else		
	cl_platform_id *platforms = NULL;
	
	// Use clGetPlatformIDs() to retrieve the number of platforms
	status = clGetPlatformIDs(0, NULL, &numPlatforms);
	cclTestSuccess2("clGetPlatformIDs", status, __LINE__);
		
	// Allocate enough space for each platform
	platforms = (cl_platform_id*)malloc(numPlatforms*sizeof(cl_platform_id));
	
	// Fill in platforms with clGetPlatformIDs()
	status = clGetPlatformIDs(numPlatforms, platforms, NULL);
	cclTestSuccess2("clGetPlatformIDs", status, __LINE__);
#endif
	
	// for each selected platform, get the selected device
	
	GLOBAL_num_devices = 0;
#ifndef _STATIC_MAX_PLATFORMS_AND_DEVICES	
	GLOBAL_dev_info = NULL;
#endif

	int p;
	for (p=0; p < numPlatforms; p++) {				
	
		//---------------------------------------------------
		// Discover and initialize the devices
		//---------------------------------------------------
	
		cl_uint numDevices = 0;
		
#ifdef _STATIC_MAX_PLATFORMS_AND_DEVICES
		cl_device_id devices[MAX_DEVICES];

		status = clGetDeviceIDs(platforms[p], CL_DEVICE_TYPE_ALL, MAX_DEVICES, devices, &numDevices);
		cclTestSuccess2("clGetDeviceIDs", status, __LINE__);
#else		
		cl_device_id *devices = NULL;
	
		// Use clGetDeviceIDs() to retrieve the number of devices present
		status = clGetDeviceIDs(platforms[p], CL_DEVICE_TYPE_ALL, 0, NULL, &numDevices);
		cclTestSuccess2("clGetDeviceIDs", status, __LINE__);	

		// Allocate enough space for each device
		devices = (cl_device_id*)malloc(numDevices*sizeof(cl_device_id));
		
		// Fill in devices with clGetDeviceIDs()
		status = clGetDeviceIDs(platforms[p], CL_DEVICE_TYPE_ALL, numDevices, devices, NULL);
		cclTestSuccess2("clGetDeviceIDs", status, __LINE__);		
#endif
	
		int d;
		for (d=0; d < numDevices; d++) {

			char param_value[255]="localhost"; size_t param_value_size_ret;
	
			status=clGetPlatformInfo(platforms[p], 0x000, 255, (void*)param_value, &param_value_size_ret);
			//cclTestSuccess2("clGetPlatformInfo", status, __LINE__);		
			printf("host\t%s\n", param_value);
			
			printf("\tplatform %d\n", p);
	
			status=clGetPlatformInfo(platforms[p], CL_PLATFORM_NAME, 255, (void*)param_value, &param_value_size_ret);
			cclTestSuccess2("clGetPlatformInfo", status, __LINE__);
			printf("\tCL_PLATFORM_NAME: %s\n", param_value);
			
#ifndef _STATIC_MAX_PLATFORMS_AND_DEVICES			
			GLOBAL_dev_info=(dev_info_t*)realloc(GLOBAL_dev_info, (GLOBAL_num_devices+1)*sizeof(dev_info_t));
#endif
			(GLOBAL_dev_info[GLOBAL_num_devices]).platform=platforms[p];
			
			printf("\t\tdevice %d\n", d);
	
			status=clGetDeviceInfo(devices[d], CL_DEVICE_NAME, 255, (void*)param_value, &param_value_size_ret);
			cclTestSuccess2("clGetDeviceInfo", status, __LINE__);
			printf("\t\t\t\tCL_DEVICE_NAME: %s\n", param_value);		
	
			(GLOBAL_dev_info[GLOBAL_num_devices]).device=devices[d];
			
			GLOBAL_num_devices++;	
			
		}
#ifndef _STATIC_MAX_PLATFORMS_AND_DEVICES	
		free(devices);
#endif
		
	}
#ifndef _STATIC_MAX_PLATFORMS_AND_DEVICES		
	free(platforms);
#endif

}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////


void get_work(int *i, int *j){
	static int ii=0, jj=0;
	
	pthread_mutex_lock(&GLOBAL_get_work_mutex);

	*i = ii;
	*j = jj;
	if((ii != -1) && (jj != -1)){
		jj += GLOBAL_slice;
		if(jj == GLOBAL_size){
			ii += GLOBAL_slice;
			jj = 0;
		}
		if(ii == GLOBAL_size)
			ii = jj = -1;
	}

	pthread_mutex_unlock(&GLOBAL_get_work_mutex);
}

////////////////////////////////////////////////////////////////////////////////

void * dev_thread_routine(void *arg){
	cl_context_properties properties[3]={CL_CONTEXT_PLATFORM, 0, 0};
	cl_context ctx;
	cl_command_queue command_queue;
	cl_mem bufferA, bufferB, bufferC;
	cl_program program;
	cl_kernel kernel;
	cl_event event[3];
	cl_int result, errcode_ret;
	cl_ulong c_start, c_end;
	size_t global_work_size[2], local_work_size[2];
	int dev, i, j, last_i=-1, last_j=-1, i2, j2;
	float *C2;

	pthread_barrier_wait(&GLOBAL_ccl_barrier);

	C2 = (float *)malloc(sizeof(float)*GLOBAL_slice*GLOBAL_slice);

	dev = (long)((int *)arg);
	properties[1] = (cl_context_properties)GLOBAL_dev_info[dev].platform;
	ctx = clCreateContext(properties, 1, &GLOBAL_dev_info[dev].device, NULL, NULL, &errcode_ret);
	cclTestSuccess("clCreateContext", errcode_ret);

	command_queue = clCreateCommandQueue(ctx, GLOBAL_dev_info[dev].device, CL_QUEUE_PROFILING_ENABLE, &errcode_ret);
	cclTestSuccess("clCreateCommandQueue", errcode_ret);

	bufferA = clCreateBuffer(ctx, CL_MEM_READ_ONLY, sizeof(float)*GLOBAL_size*GLOBAL_slice, NULL, &errcode_ret);
	cclTestSuccess("clCreateBuffer", errcode_ret);
	bufferB = clCreateBuffer(ctx, CL_MEM_READ_ONLY, sizeof(float)*GLOBAL_slice*GLOBAL_size, NULL, &errcode_ret);
	cclTestSuccess("clCreateBuffer", errcode_ret);
	bufferC = clCreateBuffer(ctx, CL_MEM_WRITE_ONLY, sizeof(float)*GLOBAL_slice*GLOBAL_slice, NULL, &errcode_ret);
	cclTestSuccess("clCreateBuffer", errcode_ret);

	program = clCreateProgramWithSource(ctx, 1, (const char **)GLOBAL_kernel_source_code, NULL, &errcode_ret);
	cclTestSuccess("clCreateProgramWithSource", errcode_ret);

	result = clBuildProgram(program, 1, &GLOBAL_dev_info[dev].device, "", NULL, NULL);
	cclTestSuccess("clBuildProgram", result);

	kernel = clCreateKernel(program, "matrix_mult", &errcode_ret);
	cclTestSuccess("clCreateKernel", errcode_ret);

	result = clSetKernelArg(kernel, 0, sizeof(int), &GLOBAL_size);
	cclTestSuccess("clSetKernelArg", result);
	result = clSetKernelArg(kernel, 1, sizeof(int), &GLOBAL_slice);
	cclTestSuccess("clSetKernelArg", result);
	result = clSetKernelArg(kernel, 2, sizeof(cl_mem), &bufferA);
	cclTestSuccess("clSetKernelArg", result);
	result = clSetKernelArg(kernel, 3, sizeof(cl_mem), &bufferB);
	cclTestSuccess("clSetKernelArg", result);
	result = clSetKernelArg(kernel, 4, sizeof(cl_mem), &bufferC);
	cclTestSuccess("clSetKernelArg", result);

	if(GLOBAL_tile == 0)
		global_work_size[0] = global_work_size[1] = local_work_size[0] = local_work_size[1] = 1;
	else{
		global_work_size[0] = global_work_size[1] = GLOBAL_slice;
		local_work_size[0] = local_work_size[1] = GLOBAL_tile;
	}
	
	get_work(&i, &j);
	while((i != -1) && (j != -1)){
		printf("[%04d-%04d][%04d-%04d]: dev=%02d, platform=%#04lx, device=%#04lx\n", i, i+GLOBAL_slice-1, j, j+GLOBAL_slice-1, dev, (long)GLOBAL_dev_info[dev].platform, (long)GLOBAL_dev_info[dev].device);
		GLOBAL_dev_info[dev].total_slices++;

		if(i != last_i){
			result = clEnqueueWriteBuffer(command_queue, bufferA, CL_FALSE, 0, sizeof(float)*GLOBAL_size*GLOBAL_slice, GLOBAL_A+GLOBAL_size*i, 0, NULL, &event[0]);
			cclTestSuccess("clEnqueueWriteBuffer", result);
			GLOBAL_dev_info[dev].total_enq_write_bytes += sizeof(float)*GLOBAL_size*GLOBAL_slice;
		}

		if(j != last_j){
			result = clEnqueueWriteBuffer(command_queue, bufferB, CL_FALSE, 0, sizeof(float)*GLOBAL_slice*GLOBAL_size, GLOBAL_B+GLOBAL_size*j, 0, NULL, &event[1]);
			cclTestSuccess("clEnqueueWriteBuffer", result);
			GLOBAL_dev_info[dev].total_enq_write_bytes += sizeof(float)*GLOBAL_size*GLOBAL_size;
		}

		result = clEnqueueNDRangeKernel(command_queue, kernel, 2, NULL, global_work_size, local_work_size, 2, event, &event[2]);
		cclTestSuccess("clEnqueueNDRangeKernel", result);

		result = clEnqueueReadBuffer(command_queue, bufferC, CL_TRUE, 0, sizeof(float)*GLOBAL_slice*GLOBAL_slice, C2, 1, &event[2], NULL);
		cclTestSuccess("clEnqueueReadBuffer", result);
		GLOBAL_dev_info[dev].total_enq_read_bytes += sizeof(float)*GLOBAL_slice*GLOBAL_slice;

		result = clGetEventProfilingInfo(event[2], CL_PROFILING_COMMAND_START, sizeof(cl_ulong), &c_start, NULL);
		cclTestSuccess("clGetEventProfilingInfo", result);
		result = clGetEventProfilingInfo(event[2], CL_PROFILING_COMMAND_END, sizeof(cl_ulong), &c_end, NULL);
		cclTestSuccess("clGetEventProfilingInfo", result);
		GLOBAL_dev_info[dev].total_kernel_work += c_end - c_start;

		for(i2=0; i2<GLOBAL_slice; i2++)
			for(j2=0; j2<GLOBAL_slice; j2++)
				GLOBAL_C[(i+i2)*GLOBAL_size+j+j2] = C2[i2*GLOBAL_slice+j2];
		
		last_i = i;
		last_j = j;
		get_work(&i, &j);
	}

	pthread_exit(NULL);
}


////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

int main (int argc, char *argv[]){
	int i, j, d, num_threads=0;
	struct timeval t0, t1;
	unsigned long long tu0, tu1;
	cl_int result;
	
	if (argc != 2 && argc != 3 && argc != 9 && (argc <11 || (argc +1) % 4 != 0)) {
		cclPrintUsage(argv);
		exit(1);
	}
	
	if (argc ==	2 || argc == 3) { // (query mode)
			if ( strcmp(argv[1], "-q") ) {
				cclPrintUsage(argv);
				exit(1);
			}
			if ( argc == 3 && (	strcmp(argv[2], "--txt") && \
								strcmp(argv[2], "--env") && \
								strcmp(argv[2], "--file") ) ) {
				cclPrintUsage(argv);
				exit(1);
			}
			cclQueryPlatformsAndDevices(argc, argv);
			exit(0);
	}
	
	// (execute mode)
	
	GLOBAL_size = atoi(argv[2]);
	GLOBAL_slice = atoi(argv[4]);
	GLOBAL_tile = atoi(argv[6]);
	
	if (argc == 9) {
		if (!strcmp(argv[7], "-f")) {
			int argc_file=0; char **argv_file=NULL;
			cclReadPlatformsAndDevicesFromFile(&argc_file, &argv_file, argv[8]);
			if (argc_file > 0) {
				printf("notice: using the following platforms and devices\n");
				//cclShowArgv(argc_file, argv_file);
				cclInitPlatformsAndDevices(argc_file, argv_file);
			}
			else {
				printf("aborting: file %s empty or malformed\n", argv[8]);
				exit(1);
			}
		}
		
		else if (!strcmp(argv[8], "all")) {
			printf("notice: using all platforms and devices\n");
			cclInitAllPlatformsAndDevices(); 
		}
		
		else {
			cclPrintUsage(argv);
			exit(1);
		}
		
	}
	else {
		printf("notice: using the following platforms and devices\n");
		//cclShowArgv(argc-7, &argv[7]);
		cclInitPlatformsAndDevices(argc-7, &argv[7]);
	}
	
	printf("notice: initialized %d <platform, device> pairs\n", GLOBAL_num_devices);
	
	GLOBAL_A = (float *)malloc(sizeof(float)*GLOBAL_size*GLOBAL_size);
	GLOBAL_B = (float *)malloc(sizeof(float)*GLOBAL_size*GLOBAL_size);
	GLOBAL_C = (float *)malloc(sizeof(float)*GLOBAL_size*GLOBAL_size);
	for(i=0; i<GLOBAL_size; i++)
		for(j=0; j<GLOBAL_size; j++){
			GLOBAL_A[i*GLOBAL_size+j] = i+1;//stored by rows
			GLOBAL_B[j*GLOBAL_size+i] = j+2;//stored by columns
		}

	gettimeofday(&t0, NULL);
	
	GLOBAL_kernel_source_code = cclReadProgramSources(1, "cl_matrix_kernel.c");
	
#ifndef _STATIC_MAX_PLATFORMS_AND_DEVICES		
	GLOBAL_dev_thread=(pthread_t *)malloc(GLOBAL_num_devices*sizeof(pthread_t*));
#endif

	num_threads=GLOBAL_num_devices;
	if(pthread_barrier_init(&GLOBAL_ccl_barrier, NULL, num_threads) != 0) {
		perror("pthread_barrier_init");
		exit(1);
	}
		
	for(d=0; d<GLOBAL_num_devices; d++){
		printf("notice: creating thread %d\n", d);
		result = clGetDeviceInfo(GLOBAL_dev_info[d].device, CL_DEVICE_PROFILING_TIMER_RESOLUTION, sizeof(size_t), &GLOBAL_dev_info[d].resolution, NULL);
		cclTestSuccess("clGetDeviceInfo", result);
		GLOBAL_dev_info[d].total_slices = GLOBAL_dev_info[d].total_enq_write_bytes = GLOBAL_dev_info[d].total_enq_read_bytes = GLOBAL_dev_info[d].total_kernel_work = 0;
		if(pthread_create(&GLOBAL_dev_thread[d], NULL, dev_thread_routine, (int *)(long)d) != 0){
			perror("pthread_create");
			exit(1);
		}
	}
	
	for(i=0; i<GLOBAL_num_devices; i++)
		pthread_join(GLOBAL_dev_thread[i], NULL);
		
	gettimeofday(&t1, NULL);
	
	tu0 = (unsigned long long)t0.tv_sec*1000000L+(unsigned long long)t0.tv_usec;
	tu1 = (unsigned long long)t1.tv_sec*1000000L+(unsigned long long)t1.tv_usec;
	printf("time=%.3lfs\n", (tu1 - tu0) / 1000000.0);
	
	for(d=0; d<GLOBAL_num_devices; d++)
		//printf("dev %02d: total_slices=%ld, total_enq_write_bytes=%ldM, total_enq_read_bytes=%ldM, total_kernel_load=%.3lfs\n", d, GLOBAL_dev_info[d].total_slices, GLOBAL_dev_info[d].total_enq_write_bytes / 1024 / 1024, GLOBAL_dev_info[d].total_enq_read_bytes / 1024 / 1024, (GLOBAL_dev_info[d].total_kernel_work * GLOBAL_dev_info[d].resolution) / 1000000000.0);
		printf("dev %02d: total_slices=%ld, total_enq_write_bytes=%ldM, total_enq_read_bytes=%ldM, total_kernel_load=%.3lfs\n", d, GLOBAL_dev_info[d].total_slices, GLOBAL_dev_info[d].total_enq_write_bytes / 1024 / 1024, GLOBAL_dev_info[d].total_enq_read_bytes / 1024 / 1024, (double)GLOBAL_dev_info[d].total_kernel_work / (double)1000000000.0);
		
	cclCheckMatrix();
	
	free(GLOBAL_A);
	free(GLOBAL_B);
	free(GLOBAL_C);
#ifndef _STATIC_MAX_PLATFORMS_AND_DEVICES
	free(GLOBAL_dev_info);
	free(GLOBAL_dev_thread);
#endif
	
	exit(0);
}
